-- PlayMate by Cyberwizzard
--
-- A player ranking system powered by other players. The ranking is based on your own ranking of a person and rankings received from other
-- players (for example guildies) or a combination of the two. When joining a party or receiving a whisper, PlayMate will inform you of any
-- information about the others (if desired).
--
-- Thanks to the testers to help work out the bugs:
-- Pandion   - Wrath of Kermit     - Bronze Dragonflight
-- Sabriella - Wrath of Kermit     - Bronze Dragonflight
-- Gotai     - Order of the Turtle - The Sha'tar

-- Todo:
-- * Fix bug in meta generator when no player rankings exist in personal storage and shared does [0.7]
-- * Add search functionality to the listing UI [0.7]
-- * Add toggle for ranking UI ordering
-- * Create a window to view details about rankings shown in the listing UI [0.7]
-- * Polish UI and perhaps even intergrate with social panel [0.8]
-- * Add minimap icon [0.8]
-- * Add titan bar hooks [0.8]
-- * Add tracking of users that use the plugin (for debugging) [0.9]
-- * Remove beta status after testing [1.0]

-- Log:
-- Version 0.6:
-- * Patch 4.0.1 compliant
-- * New UI panel created as main frame
-- * Revamped usage frame to scroll and embed in the main UI
-- * Added about tab to main UI to show credits
-- * Created utility functions to compose, search and reorder the listings for the GUI based on the meta-database
-- * Created scrolling frame to display the player rankings using the new listing functions (incomplete)
-- * Support network protocol numbering. It will now be possible to sync between multiple PlayMate versions until protocol changes
--   actually break compatibility. Until now, each release was incompatible with an earlier version.
-- Version 0.5:
-- * Fixed bug when sharing rankings (prevented synchronization)
-- * Fixed bug when using PlayMate and Stubby (the latter uses common names globally)
-- * Added hooks to rank random players from the popup menu
-- * Fixed hook to rank players from the chat window
-- * Improved communication debugging
-- * Internal version number not increased to be compatible with 0.4, will be fixed in next release
-- Version 0.4:
-- * Added support for multi-realm ranking
-- * Rewrite of settings for multi-realm handling
-- * Added support to extract realm names from combined labels
-- * Improved handling of shared rankings
-- * Added GUI for help
-- * Fixed issues with guild and party sharing while not in one
-- * Fixed potential error when API provides incorrect unit info
-- * Fixed player lookups when the target realm is not present in the database
-- Version 0.3:
-- * Added communication throttling to prevent trashing hangs
-- * Fixed numerous bugs and improved ranking calculations
-- Version 0.2:
-- * Tested internally
-- * Shared rankings
-- * Settings for sharing rules
-- Version 0.1:
-- * Tested internally
-- * Allow ranking of other players
-- * Allow adding noted to rankings
-- * Added hooks to tooltips, chat and party/raid to use the rankings

SLASH_PLAYMATE1, SLASH_PLAYMATE2 = '/pm', '/playmate'; -- Allow the usage of /pm and /playmate to invoke functions

-- Define the PlayMate object
PlayMate = {}

-- Set the version number and protocol number. The protocol number is simply increased once the protocol changes.
PlayMate.Version = 0.6
PlayMate.Protocol = 0.4

-- Set to true to print information about the inner workings and function calls
PlayMate.debug = false

-- Pointer to the default frame to hook events
PlayMate.Frame = nil
-- PlayMate settings table, empt until the settings get loaded
PlayMate.Settings = nil

-- Use a predefined color schema
-- Use FF0000 to FF9900 for bad (red to orange) and FFFF00 to 00FF00 for good (yellow to green)
PlayMate.ColorMap = {
	["-5"] = "FFFF0000",
	["-4"] = "FFFF2200",
	["-3"] = "FFFF4400",
	["-2"] = "FFFF6600",
	["-1"] = "FFFF8800",
	["0"] =  "FFFFFF00",
	["1"] =  "FFCCFF00",
	["2"] =  "FF98FF00",
	["3"] =  "FF64FF00",
	["4"] =  "FF30FF00",
	["5"] =  "FF00FF00",
	["Text1"] = "FFBBBB00",
	["Text2"] = "FFCCCCCC",
	["Text3"] = "FF22BBBB"
}

PlayMate.HexMap = {
	["-5"] = {1.0,0.000,0.0},
	["-4"] = {1.0,0.133,0.0},
	["-3"] = {1.0,0.266,0.0},
	["-2"] = {1.0,0.399,0.0},
	["-1"] = {1.0,0.533,0.0},
	["0"] =  {1.0,1.0,0.0},
	["1"] =  {0.8,1.0,0.0},
	["2"] =  {0.6,1.0,0.0},
	["3"] =  {0.4,1.0,0.0},
	["4"] =  {0.2,1.0,0.0},
	["5"] =  {0.0,1.0,0.0}
}

PlayMate.RepNames = {
	["-5"] = "Very Bad",
	["-4"] = "Very Bad",
	["-3"] = "Bad",
	["-2"] = "Poor",
	["-1"] = "Low",
	["0"] =  "Neutral",
	["1"] =  "Good",
	["2"] =  "Nice",
	["3"] =  "Very Nice",
	["4"] =  "Great",
	["5"] =  "Awesome",
}

-- Will be loaded when the settings are loaded
PlayMate.PlayerName = "Unknown"
PlayMate.Realm = "Unknown"

-- Used to track changes in party attendance so we only report on new ppl (also used for raids)
PlayMate.Party = {}

-- Activate the commands for the console
function SlashCmdList.PLAYMATE(full_cmd)
	cmd, rest = full_cmd:match("^(%S*)%s*(.-)$");

	if cmd == "list" then
		-- List all players in the database
		PlayMate:listPlayers(rest)
	elseif cmd == "add" then
	   -- Add or replace a player ranking in the database
	   PlayMate:addPlayerCmd(rest)
	elseif cmd == "remove" then
		-- Remove a player ranking in the database
	   PlayMate:removePlayer(rest)
	elseif cmd == "rebuild" then
		-- Rebuild the metadatabase
		PlayMate:rebuildMetadata()
	elseif cmd == "party" then
		-- Show the party or raid report
		PlayMate:rankParty(true)
	elseif cmd == "sync" then
		-- Ping the guild, everyone will reply with a pong which will trigger a sync
		PlayMate:broadcastPing()
	elseif cmd == "clean" then
		-- Revert to default settings
		print("PlayMate: Reverting to default settings")
		PlayMate:loadDefaultSettings()
	elseif cmd == "purge" then
		-- Purge shared rankings
		print("PlayMate: Purging shared rankings...")
		PlayMate.Settings.Shared = {}
		-- Rebuild the metadata
		PlayMate:rebuildMetadata()
	elseif cmd == "debug" then
		-- Toggle debug
		if PlayMate.debug then
			print("PlayMate: disabling debug mode")
			PlayMate.debug = false
		else
			print("PlayMate: enabling debug mode")
			PlayMate.debug = true
		end
	elseif cmd == "show" then
		-- Show the main UI
		PlayMate:ShowUI();
	else
		-- Print usage information
		print("PlayMate version " .. PlayMate.Version .. " by Cyberwizzard -- showing usage screen")

		PlayMate:ShowUsageUI()
	end
end

-- ================================================================
-- ==                           Utils                            ==
-- ================================================================

-- Since lua tables are only holding counters for numerical indices, we need a helper function to count
-- elements otherwise....
function tcount(tab)
	local n = #tab
	if (n == 0) then
		for _ in pairs(tab) do
			n = n + 1
		end
	end
	return n
end

-- Clone table content
function tcopy(tab)
	local tab2 = {}
	for key, value in pairs(tab) do
		tab2[key] = value
	end
	return tab2
end

-- Encodes a table into a string holding key and value pairs delimited by a semi-colon (;) - if a ; is found in the value, it is escaped.
function PlayMate:encodeTable( table )
	msg = ""
	first = true
	for key, value in pairs(table) do
		-- Escape any ; in a field value
		value = tostring(value):gsub(";", "\\;")
		-- Apppend the two and prefix a ; if this is not the first value
		if first == false then msg = msg .. ";" end
		msg = msg .. key .. "=" .. value
		-- Mark the first record
		first = false
	end
	
	-- Return the constructed message
	return msg
end

-- Decodes a message received from another PlayMate instance, this is a ; separated table. Note that a ; in a value will be escaped
function PlayMate:decodeTable( msg )
	raw = {}			-- Table holding decoded values
	hasPart = true		-- True while more keys and values have not yet been decoded
	offset = 0			-- Offset to use when an escaped ; is found to skip over it while parsing
	sanity = 0			-- Sanity counter while decoding, when over 50 parts are found the function fails

	while hasPart do
		-- Find the index of the next ; in the string
		nextPart = msg:find(";", offset)
		
		-- If no ; is found anymore, abort
		if nextPart == nil then 
			hasPart = false
			-- Set the marker to the end of the string
			nextPart = msg:len()
		end
		
		-- Check to make sure this is not and escaped one
		if msg:sub( nextPart-1, nextPart ) == "\\;" then
			-- This is an escaped ;, set the offset and skip ahead
			offset = nextPart + 1
		else
			-- This is not an escaped ;, reset the offset to zero
			offset = 0
			-- Now split this into a key and a value pair
			key, value = msg:sub(1, nextPart):match("^([^=]+)=(.-);?$")
			
			-- Safety: make sure this is valid
			assert( key ~= nil, "Decoded table key is nil" )
			assert( value ~= nil, "Decoded table value is nil" )
			
			-- Don't forget to replace \; with ;
			value = value:gsub("\\;", ";")
			
			-- Add to the table
			raw[ key ] = value
			
			-- Cut the string off the beginning
			msg = msg:sub( nextPart+1, msg:len() )
		end
		
		-- Sanity fail-safe to prevent infinite loops
		sanity = sanity + 1
		assert(sanity < 50, "decodeTable had more than 50 iterations - aborted")
	end
	
	-- Return the decoded table
	return raw
end

local bit_band, bit_bxor, bit_rshift, str_byte, str_len = bit.band, bit.bxor, bit.rshift, string.byte, string.len
local consts = { 0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA, 0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3, 0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988, 0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91, 0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE, 0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7, 0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC, 0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5, 0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172, 0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B, 0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940, 0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59, 0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116, 0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F, 0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924, 0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D, 0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A, 0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433, 0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818, 0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01, 0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E, 0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457, 0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C, 0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65, 0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2, 0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB, 0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0, 0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9, 0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086, 0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F, 0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4, 0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD, 0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A, 0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683, 0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8, 0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1, 0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE, 0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7, 0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC, 0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5, 0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252, 0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B, 0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60, 0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79, 0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236, 0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F, 0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04, 0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D, 0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A, 0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713, 0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38, 0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21, 0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E, 0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777, 0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C, 0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45, 0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2, 0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB, 0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0, 0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9, 0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6, 0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF, 0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94, 0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D }
 
function PlayMate:CRC32(s)
	local crc, l, i = 0xFFFFFFFF, str_len(s)
	for i = 1, l, 1 do
		crc = bit_bxor(bit_rshift(crc, 8), consts[bit_band(bit_bxor(crc, str_byte(s, i)), 0xFF) + 1])
	end
	return bit_bxor(crc, -1)
end

-- Searches the character name for a dash - this is the divider to seperate the name from the realm name. If no dash is found,
-- the local realm name is returned. This function returns two values: name and realm
function PlayMate:decodeCharacterName( characterName )
	name = ""
	realm = ""
	
	-- Find the dash in the name (if present)
	dash = characterName:find("-")
	
	if dash == nil then
		-- No dash found, just return the current name and realm
		return characterName, PlayMate.Realm
	end
	
	name = characterName:sub(1, dash-1)
	realm = characterName:sub(dash+1, characterName:len())

	return name, realm
end

function PlayMate:countLocalRankings()
	count = 0
	for realm, playerList in pairs(PlayMate.Settings.Mine) do
		count = count + tcount(playerList)
	end
	return count
end


-- ================================================================
-- ==                            Core                            ==
-- ================================================================

-- Handle the load of the plugin. The init generates the frame and sets up the event handlers but not the settings and data.
-- When the addon is fully loaded and the persistent data is loaded as well, we try to load the data and process it if needed.
function PlayMate:loadSettings() 
	-- Only allow the settings load to be run once
	if not self.Loaded then
		self.Loaded = 1
	else
		return
	end
	
	-- Spam the load
	print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate " .. PlayMate.Version .. " loaded -|r |c" .. PlayMate.ColorMap["Text2"] .. " type /pm or /playmate to show help|r")
	
	-- Determine the current user (important to filter out messages originating from us)
	PlayMate.PlayerName, dummy = UnitName("player")
	PlayMate.Realm = GetRealmName()
	
	-- Attempt to load the settings from the global scope
	PlayMate.Settings = getglobal("PlayMateSettings")
	
	-- Test if the settings are valid, if not force the load of the default settings
	if PlayMate.Settings == nil then PlayMate:loadDefaultSettings() end
	
	-- Temporary fix: everyone using 0.3 will not have a UUID loaded, generate one now
	if PlayMate.Settings["UUID"] == nil then PlayMate.Settings["UUID"] = PlayMate:CRC32(PlayMate.PlayerName) end
	
	-- Test if the version is correct
	if PlayMate.Settings.Version ~= 2 then
		if PlayMate.Settings.Version == 1 then
			print("Upgrading PlayMate settings from version 1 to version 2");
			-- Version 1 had no realm info for rankings, add them now (for lack of data, assume all are local)
			for name,playerObj in pairs(PlayMate.Settings["Mine"]) do
				PlayMate.Settings["Mine"][name]["Realm"] = PlayMate.Realm
			end
			-- Now move all local rankings in a dedicated store, sorted by realm
			backup = PlayMate.Settings["Mine"]
			-- Wipe the old store
			PlayMate.Settings["Mine"] = {}
			-- Add the rankings in the array element by the instance name
			PlayMate.Settings["Mine"][PlayMate.Realm:lower()] = backup
			
			-- Also add realm info to the shared data
			for UUID, data in pairs(PlayMate.Settings.Shared) do
				for playerKey, playerObj in pairs( data["Data"] ) do
					-- Set the realm
					PlayMate.Settings.Shared[ UUID ]["Data"][ playerKey ]["Realm"] = PlayMate.Realm
				end
				-- Now copy all rankings from the current remote account into a new array using the realm name
				local backup = PlayMate.Settings.Shared[ UUID ]["Data"]
				-- Wipe the storage
				PlayMate.Settings.Shared[ UUID ]["Data"] = {}
				-- Insert data again
				PlayMate.Settings.Shared[ UUID ]["Data"][ PlayMate.Realm:lower() ] = backup
			end
			
			PlayMate.Settings.Version = 2
			PlayMate:saveSettings()
			
			-- Mark the settings as being dirty (to force a metadata rebuild)
			PlayMate.Settings.Dirty = true
		else
			print("PlayMate: Incompatible settings found, reverting to defaults")
			PlayMate:loadDefaultSettings()
		end
	end
	
	-- If the rankings are dirty, we need to rebuild the metadata list
	if PlayMate.Settings.Dirty then
		PlayMate:rebuildMetadata()
	end
	
	-- Add the player name to the table of personal character names: if we receive rankings for those names
	-- or try to rank those ourselves, we will ignore those rankings
	-- temp fix: remove as soon as all the testers have this version
	if PlayMate.Settings.MyCharacters == nil then PlayMate.Settings.MyCharacters = {} end
	local currentPlayer = PlayMate.PlayerName:lower()
	if PlayMate.Settings.MyCharacters[ currentPlayer ] == nil then
		-- First time this character has been logged on, add to the list
		print("PlayMate: detected new character " .. PlayMate.PlayerName)
		PlayMate.Settings.MyCharacters[ currentPlayer ] = time()
		
		-- Keep track if we deleted rankings from the personal or shared rankings
		local found = false
		
		-- Check the personal storage for rankings of your own characters
		if PlayMate.Settings.Mine[ currentPlayer ] ~= nil then
			print("Character name " .. PlayMate.PlayerName .. " found in your rankings, removing...")
			PlayMate:removePlayer( currentPlayer )
			found = true
		end
		
		-- Search all shared rankings for this character name
		for UUID, data in pairs(PlayMate.Settings.Shared) do
			for playerKey, playerObj in pairs( data["Data"] ) do
				if playerKey == currentPlayer then
					-- Found a ranking belonging to the newly logged in character, remove it
					PlayMate.Settings.Shared[ UUID ]["Data"][ playerKey ] = nil
					found = true
				end
			end
		end
		
		-- If we found this character name in any rankings (personal or shared), rebuild the db as the data will be dirty
		if found then PlayMate:rebuildMetadata() end
	end
	
	-- Ping the guild for updates as we log in and get ready to play
	PlayMate:broadcastPing()
end

-- When the loadSettings function detects the total lack of settings, this function
-- will init the settings table with the minimal set of values.
function PlayMate:loadDefaultSettings() 
	if PlayMate.debug then print('PlayMate: loading default settings') end
	
	-- If an old UUID was set, make sure it survives a reset of all data
	local oldUUID = nil
	if PlayMate.Settings ~= nil and PlayMate.Settings["UUID"] ~= nil then
		oldUUID = PlayMate.Settings["UUID"]
	end

	PlayMate.Settings = {}
	-- Set the version info
	PlayMate.Settings.Version = 2
	-- This is the list of player rankings created by me (or imported into my own notes)
	PlayMate.Settings.Mine = {}
	-- This is the list of shared (imported) rankings. Each index is a player name who sent it to us and contains a
	-- table holding PlayerName, lastSync and Data. Data holds a player ranking table like Settings.Mine.
	PlayMate.Settings.Shared = {}
	
	-- The MyCharacters table holds all the characters that have been logged in on this realm and account. That way we
	-- keep track of alts and make sure everyone syncs with one account rather than one characters (preventing a ranking
	-- bash or boost).
	PlayMate.Settings.MyCharacters = {}
	
	-- This table holds the combined data from Mine and Shared into one lookup table holding shared rankings. Note that
	-- the Rank value is not an integer and that Votes tells us how many people have voted for this ranking.
	PlayMate.Settings.Players = {}
	-- Receiving ranking updates from other players will not trigger a metadata rebuild on its own
	-- Instead, Dirty is set to true and upon the next login, the metadata will be rebuild.
	PlayMate.Settings.Dirty = false
	
	-- Reporting settings - boolean flags
	PlayMate.Settings.ShowRaidRankings = true				-- Show ranking reports in a raid when the raid changes
	PlayMate.Settings.ShowPartyRankings = true			-- Show ranking reports in a party when the party changes
	PlayMate.Settings.ShowInLFG = true							-- Show ranking when people talk in LFG
	PlayMate.Settings.ShowInWhispers = true					-- Show ranking when receiving whispers
	PlayMate.Settings.ShareParty = true							-- Share rankings in the party
	PlayMate.Settings.ShareGuild = true							-- Share rankings in the guild
	
	-- Create a CRC32 hash from the current user name if the old one is not set
	if oldUUID == nil then
		PlayMate.Settings["UUID"] = PlayMate:CRC32(PlayMate.PlayerName)	
	end
	-- Save the new settings
	PlayMate:saveSettings()
end

-- Save the current settings into the global storage
function PlayMate:saveSettings()
	if PlayMate.debug then print('PlayMate: saving current settings') end
	setglobal("PlayMateSettings", PlayMate.Settings)
end

-- Global init, gets run after most parts of PlayMate have been loaded
function PlayMate:Init()
	if PlayMate.debug then print('PlayMate: running init') end

	-- Generate a hidden frame to grab events
	self.Frame = CreateFrame("Frame")
	
	-- Activate the events we need to monitor
	self.Frame:RegisterEvent("ADDON_LOADED")					-- Used to load settings
	self.Frame:RegisterEvent("PARTY_MEMBERS_CHANGED")			-- Used to monitor party events and report on members
	self.Frame:RegisterEvent("PLAYER_LEAVING_WORLD")			-- Used to trigger a save of the settings
	self.Frame:RegisterEvent("CHAT_MSG_ADDON")					-- Used to received messages from other players
	self.Frame:RegisterEvent("PLAYER_ENTERING_WORLD")			-- Used for rebuilding metadata when rankings have been received from others
	
	-- Set the function to handle the events
	self.Frame:SetScript("OnEvent", self.handleEvent)
end


-- Find a player from both your player ranking and the shared ones. The returned player object holds
-- a ranking based on both rankings. Note that the 'realm' is optional but defaults to the local realm.
-- Returns nil when no user is found with the given name
function PlayMate:searchPlayer(name, realm)
	-- Sanity checking
	if name == nil then return nil end
	if realm == nil then realm = PlayMate.Realm end
	-- If for some reason someone succeeds to start a search with an empty string, assume local realm
	if realm:trim():len() == 0 then realm = PlayMate.Realm end
	
	-- Make sure the realm is present in the player rankings (thanks to Gotai for catching this)
	if PlayMate.Settings.Players[realm:lower()] == nil then return nil end
	
	-- Look the player up and return it, when the player does not exist, a nil is returned
	return PlayMate.Settings.Players[realm:lower()][name:lower()]
end

-- This function searches the personal and shared (aka received from others) rankings for a given player name.
-- Note that a realm can be added by providing the 'name-realm' format, otherwise it defaults to the current realm.
function PlayMate:listPlayers(searchname)
	searchrealm = ""
	if searchname ~= "" then
		searchname, searchrealm = PlayMate:decodeCharacterName(searchname)
	end
	-- Convert to lower case
	searchrealm_l = searchrealm:lower()

	-- Show what we are doing
	if searchname == "" then
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate listing all players in database")
		-- Erase the realm info, if any to match all realms
		searchrealm = ""
		searchrealm_l = ""
	else
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate listing players in database |c" .. PlayMate.ColorMap["Text2"] .. " matching '" .. searchname .. "' from '" .. searchrealm .. "'")
	end
	
	-- List your own list
	local found = 0
	local headerPrinted = false
	-- Iterate over all realms
	-- Todo: sort the results into an alphabetical something?
	for realm,list in pairs(PlayMate.Settings["Mine"]) do
		-- Check if the realm name matches, if no name was provided, this will be empty as we want to list all
		if realm == searchrealm_l or searchrealm_l == "" then
			-- Realm name matched
			for name,playerObj in pairs(list) do
				if searchname == "" or name:lower():find( searchname:lower() ) then
					-- Found one! Check if we need to print the header
					if found == headerPrinted then
						print("Players ranked by you: (Name: Score)");
						headerPrinted = true
					end
				
					local rankStr = PlayMate:getColoredRanking(playerObj["Rank"])
					local noteStr = "Note: "
					if playerObj["ShareNote"] == false then noteStr = "Private Note: " end
					print("   " .. playerObj["PlayerName"] .. "-" .. playerObj["Realm"] .. ": " .. rankStr .. " - " .. noteStr .. playerObj["Note"])
					
					-- Count the number of matches
					found = found + 1
				end
			end
		end
	end
	
	-- List the shared listings
	for UUID, sharedObj in pairs(PlayMate.Settings["Shared"]) do
		headerPrinted = false	-- Flag if the header has been printed for this table
		for realm, playerList in pairs( sharedObj["Data"] ) do
			if realm == searchrealm_l or searchrealm_l == "" then
				for name, playerObj in pairs( playerList ) do
					if searchname == "" or name:lower():find( searchname:lower() ) then
						-- If no data has been found before for this players rankings, print the header
						if headerPrinted == false then
							-- Print header
							print("Players ranked by " .. sharedObj["PlayerName"] .. ": (Name: Score - Note)");
							-- Toggle it
							headerPrinted = true
						end
					
						local rank = playerObj["Rank"]
						print("   " .. playerObj["PlayerName"] .. "-" .. playerObj["Realm"] .. ": " .. PlayMate:getColoredRanking(rank) .. " - Note: " .. playerObj["Note"])
						
						-- Count number of matches
						found = found + 1
					end
				end
			end
		end
	end

	-- Wrap up: warn if not found
	if found == false then
		-- Create warning
		local str = "|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: |r|c" .. PlayMate.ColorMap["Text2"] .. "0 players found"
		
		-- Add the search string to the message if set
		if searchname ~= nil and searchname:len() > 0 then
			str = str .. " containing |r'" .. searchname .. "' from realm '" .. searchrealm .. "'"
		end
		
		print(str)
	else
		-- Create info message
		local str = "|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: |r|c" .. PlayMate.ColorMap["Text2"] .. found .." player rankings found"
		print(str)
	end
end

function PlayMate:getRankColor(ranking)
	-- Convert to a number to be sure
	rank = tonumber(ranking, 10)
	-- If not a valid number, set it to neutral
	if rank == nil then rank = 0 end

	-- Make sure it is in the -5 to 5 bounds
	if rank < -5 then rank = -5 end
	if rank > 5 then rank = 5 end

	-- Round it if needed
	rank = floor(rank)

	-- Return the color code
	return PlayMate.ColorMap[tostring(rank)]
end

-- Get a string detailing the ranking in title, ranking value and color
function PlayMate:getColoredRanking(ranking)
	-- Convert to a number to be sure
	rank = tonumber(ranking, 10)
	-- If not a valid number, set it to neutral
	if rank == nil then rank = 0 end

	-- Make sure it is in the -5 to 5 bounds
	if rank < -5 then rank = -5 end
	if rank > 5 then rank = 5 end
	
	-- Save the unrounded value in 'ranking', this fixes an error when ranking is nil (and gets concatted below)
	ranking = rank

	-- Round it if needed
	rank = tostring( floor(rank) )

	-- Return the color code
	return "|c" .. PlayMate.ColorMap[ rank ] .. PlayMate.RepNames[ rank ] .. " (" .. ranking .. ")|r"
end

-- Get the rank string
function PlayMate:getRankName(ranking)
	-- Convert to a number to be sure
	rank = tonumber(ranking, 10)
	-- If not a valid number, set it to neutral
	if rank == nil then rank = 0 end

	-- Make sure it is in the -5 to 5 bounds
	if rank < -5 then rank = -5 end
	if rank > 5 then rank = 5 end

	-- Round it if needed
	rank = tostring( floor(rank) )
	
	return PlayMate.RepNames[ rank ]
end

-- Called when the console gets a /pm add
function PlayMate:addPlayerCmd(args)
	name, rank, note = args:match("^(%S*)%s*(\-?[0-9]+)%s*(.-)$")
	
	if name == nil or name == "" then
		--print("PlayMate usage: /pm add PlayerName Rank (Description)")
		PlayMate:showAddForm()
		return
	end
	
	-- Add the player using the extracted values
	PlayMate:addPlayer(name, tonumber(rank), note, true, true)
end

-- This function takes the properties to rank a player and add the player to the local database
-- if everything is valid. If desired, the ranking is sent to others afterwards.
function PlayMate:addPlayer(name, rank, note, share, broadcast)
	-- To make everything idiot-proof (thanks to Sabriella for proving we needed this), apply some
	-- conditional checks on everything

	-- Make sure the values are valid
	assert(name ~= nil, "PlayMate addPlayer error: name is nil!")
	assert(type(name) == "string", "PlayMate addPlayer error: name is not a string!")
	assert(name:trim() ~= "", "PlayMate addPlayer error: name is empty!")
	
	assert(type(rank) == "number", "PlayMate addPlayer error: rank is not a number!")
	assert(rank >= -5 and rank <= 5, "PlayMate addPlayer error: rank is out of range! [-5 .. 5]")
	
	-- Convert a nil note to an empty string and any other type to an empty string
	if type(note) ~= "string" then note = "" end
	if note == nil then note = "" end
	
	-- Make sure the share and broadcast vars are boolean
	if share ~= true and share ~= false then share = true end
	if broadcast ~= true and broadcast ~= false then broadcast = true end
	
	-- Split the name using the dash in a character name and a realm name
	name, realm = PlayMate:decodeCharacterName(name)
	-- Convert Realm to lower case as we use it as an index
	realm_l = realm:lower()

	-- Create a lowercase index from the name
	name_l = name:lower()	-- Convert to lower case for index usage
	
	-- Make sure rank is an integer number
	rank = floor(tonumber(rank, 10))
	
	-- Make sure it is in range
	-- Note: after the assert above this is probably moot
	if rank < -5 or rank > 5 then
		print("PlayMate error: a valid rank is a whole number from -5 to 5 and not " .. rank)
		return false
	end
	
	-- Make sure the note is less than 128 chars long, this saves bandwidth while syncing and should be more than enough
	if note:len() > 128 then
		print("PlayMate error: a note can only hold 128 characters, " .. note:len() .. " given")
		return false
	end
	
	-- Check if the ranking is about a character owned by this account
	if PlayMate.Settings.MyCharacters[ name_l ] ~= nil then
		print("PlayMate: you are not allowed to rank your own characters (for obvious reasons) - ranking denied")
		return false
	end

	-- Add or replace?
	if PlayMate.Settings.Mine[ realm_l ] == nil then
		PlayMate.Settings.Mine[ realm_l ] = {}
	end
	if PlayMate.Settings.Mine[ realm_l ][ name_l ] == nil then
		-- Create a new record
		PlayMate.Settings.Mine[ realm_l ][ name_l ] = {
			["PlayerName"] = name,
			["Realm"] = realm,
			["Rank"] = tostring(rank),
			["Note"] = note,
			["ShareNote"] = share,
			["Modified"] = time(),
			["Created"] = time()
		}
		
		-- Tell the user what we did
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: added player '" .. name .. "' from '" .. realm .. "' to database with rank " .. rank .. " and note: |r'" .. note .. "'")
	else
		-- Update an existing record
		PlayMate.Settings.Mine[realm_l][name_l]["Rank"] = tostring(rank)
		PlayMate.Settings.Mine[realm_l][name_l]["Note"] = note
		PlayMate.Settings.Mine[realm_l][name_l]["ShareNote"] = share
		
		-- Update the timestamp
		PlayMate.Settings.Mine[realm_l][name_l]["Modified"] = time()
		
		-- Tell the user what went down
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: updated player '" .. name .. "' from '" .. realm .. "' to have rank " .. rank .. " and note: |r'" .. note .. "'")
	end
	
	-- Rebuild the metadata
	PlayMate:rebuildMetadata()
	
	-- If for some reason we shouldn't broadcast this ranking at all, abort now
	if broadcast == false then return true end

	-- Send the player ranking to others
	if PlayMate.Settings.ShareGuild == true and IsInGuild() then
		if PlayMate.debug then
			print("Sending ranking to guild...")
		end
		PlayMate:broadcastPlayerRanking( PlayMate.Settings.Mine[realm_l][name_l], nil, "GUILD" )
	end
	
	if PlayMate.Settings.ShareParty == true and GetNumPartyMembers() > 0 then
		if PlayMate.debug then
			print("Sending ranking to party...")
		end
		PlayMate:broadcastPlayerRanking( PlayMate.Settings.Mine[realm_l][name_l], nil, "PARTY" )
	end
	
	if PlayMate.debug then
		print("Ranking complete")
	end
	
	-- After broadcasting, let it be known that is was all ok
	return true
end

-- Remove a player from your own ranking list
function PlayMate:removePlayer(name)
	-- Split the name in a character and realm
	name, realm = PlayMate:decodeCharacterName(name)

	-- Convert to lower case
	name_l = name:lower()
	-- Convert realm as well
	realm_l = realm:lower()

	if PlayMate.Settings.Mine[realm_l][ name_l ] ~= nil then
		PlayMate.Settings.Mine[realm_l][ name_l ] = nil
		
		-- Inform the user
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: removed player '" .. name .. "' from '" .. realm .. "' from database")
	else
		-- Unknown user
		print("|c" .. PlayMate.ColorMap["Text1"] .. "PlayMate: have no records for player '" .. name .. "' from '" .. realm .. "' to remove")
	end
end

-- This function grabs the shared player rankings from everyone and generates an average score.
-- This score is then averaged with your personal score, if any for that player.
function PlayMate:rebuildMetadata()
	print("PlayMate: Rebuilding metadata for all rankings...")

	-- Destroy the old data
	PlayMate.Settings.Players = {}
	
	-- Insert all players from your own database
	for realmName, playerList in pairs(PlayMate.Settings.Mine) do
		PlayMate.Settings.Players[realmName] = {}
		for playerName, playerObj in pairs(playerList) do
			-- Copy all relevant data
			PlayMate.Settings.Players[realmName][playerName] = {
				["PlayerName"] = playerObj["PlayerName"],
				["Rank"] = playerObj["Rank"],
				["Note"] = playerObj["Note"],
				["Votes"] = 0,
				["TotalRank"] = 0
			}
		end
	end
	
	-- Loop over all shared rankings
	for UUID, sharedObj in pairs(PlayMate.Settings.Shared) do
		for realm, playerList in pairs(sharedObj["Data"]) do
			-- Add the realm if its not yet added
			if PlayMate.Settings.Players[realm] == nil then
				PlayMate.Settings.Players[realm] = {}
			end
			
			-- Now add the player ranking to the realm list
			for playerName, playerObj in pairs(playerList) do
				-- Make sure the player is known
				if PlayMate.Settings.Players[realm][playerName] == nil then
					-- Add it, note that we ommit Rank here as a missing local ranking means we have no local rank
					PlayMate.Settings.Players[realm][playerName] = {
						["PlayerName"] = playerName,
						["TotalRank"] = 0,
						["Votes"] = 0,
						["Note"] = playerObj["Note"]
					}
				end
				
				-- Make sure the shared counters exist
				if PlayMate.Settings.Players[realm][playerName]["TotalRank"] == nil then
					PlayMate.Settings.Players[realm][playerName]["TotalRank"] = 0
					PlayMate.Settings.Players[realm][playerName]["Votes"] = 0
				end
				
				--print("Debug: ", playerName, " Rank: ", PlayMate.Settings.Players[realm][ playerName ]["TotalRank"])
				
				-- Add the rank to the counters
				PlayMate.Settings.Players[realm][playerName]["TotalRank"] = PlayMate.Settings.Players[realm][playerName]["TotalRank"] + playerObj["Rank"]
				PlayMate.Settings.Players[realm][playerName]["Votes"] = PlayMate.Settings.Players[realm][playerName]["Votes"] + 1
			end
		end
	end

	-- Now consolidate into a single ranking and remove other data
	for realmName, playerList in pairs(PlayMate.Settings.Players) do
		for playerName, playerObj in pairs(playerList) do
			-- Fetch the combined totals, if set
			local totalRank = 0
			if playerObj["TotalRank"] ~= nil then totalRank = playerObj["TotalRank"] end
			local votes = 0
			if playerObj["Votes"] ~= nil then votes = playerObj["Votes"] end
			
			-- Generate the combined ranking and set it to the total rank, if we have any shared votes
			local rank = nil
			-- We have votes, calculate the average rank
			if votes ~= 0 then rank = totalRank / votes end
			
			-- If we have a local ranking, combine that with the current (shared average rank)
			if playerObj["Rank"] ~= nil then
				-- Add the number of votes as we votes ourselves
				votes = votes + 1
				-- Average between our vote and the common consensus, if there was a shared ranking
				if rank ~= nil then
					-- We have a shared ranking in 'rank', combine it with the local rank
					rank = (rank + playerObj["Rank"]) / 2
					--print("Shared: ", sharedRank, " My: ", playerObj["Rank"], " Avg: ", rank)
				else
					-- Rank is nil which means there was no shared ranking for this player.
					-- Note that in the case that there is no shared ranking, the local rank remains as copied from the
					-- local database.
					rank = playerObj["Rank"]
				end
				
				-- Save the number of votes after correction
				playerObj["Votes"] = votes
			end
			
			-- Safety test
			assert(rank ~= nil, "Rank is nil when combining rankings, this is impossible!")
	
			-- Round to 1 digit
			rank = floor(rank * 10) / 10
			
			-- Replace the ranking with the combined value
			playerObj["Rank"] = rank
			
			-- Unset the temp values
			playerObj["TotalRank"] = nil
		end
	end
	
	-- Mark the rebuild
	PlayMate.Settings.Dirty = false
	
	-- Save the settings
	PlayMate:saveSettings()
	
	-- Count the complete number of rankings
	local count = 0
	local rcount = 0
	for realm, list in pairs(PlayMate.Settings.Players) do
		count = count + tcount(list)
		rcount = rcount + 1
	end
	
	if rcount == 1 then
		print("PlayMate: Rebuild complete for " .. count .. " players in 1 realm")
	else
		print("PlayMate: Rebuild complete for " .. count .. " players in " .. rcount .. "realms")
	end
end

-- This function is called every time the party or raid changed. To prevent spamming we only
-- show player rankings we haven't shown before in this party.
function PlayMate:rankParty(force)
	if force == nil then 
		force = false
	end
	
	-- If we force a full party/raid report, simply wipe the attendance list
	if force then
		PlayMate.Party = {}
	end
	
	if GetNumRaidMembers() > 0 then
		-- This is a raid, abort if we dont want this
		if PlayMate.Settings.ShowRaidRankings == false then return end
		
		printed = false
		for index = 1, GetNumRaidMembers() do
			name, realm = UnitName("raid" .. index)
			
			-- See if we have reported on this member before (prevent spamming)
			if PlayMate.Party[ name ] == nil then
				-- Fetch any information about this member (note: raids are always local realm)
				playerObj = PlayMate:searchPlayer( name )
				
				-- If we found something, print the information
				if playerObj ~= nil then
					if printed == false then
						print("PlayMate Raid Reputation Report:")
					end
					
					-- Print the player ranking
					print("Player '" .. name .. "': " .. PlayMate:getColoredRanking( playerObj["Rank"] ))
					
					-- Toggle the flag
					printed = true
				end
				
				-- Add this member to the ignore/party list
				PlayMate.Party[ name ] = true
			end
		end
	elseif GetNumPartyMembers() > 0 then
		-- This is a party, abort if we dont want this
		if PlayMate.Settings.ShowPartyRankings == false then return end
		
		printed = false
		for index = 1, GetNumPartyMembers() do
			name, realm = UnitName("party" .. index)
			
			-- See if we have reported on this member before (prevent spamming)
			if PlayMate.Party[ name ] == nil then
				-- Fetch any information about this member
				playerObj = PlayMate:searchPlayer(name, realm)
				
				-- If we found something, print the information
				if playerObj ~= nil then
					if printed == false then
						print("PlayMate Party Reputation Report:")
					end
					
					-- Print the player ranking
					if realm == nil or realm:trim():len() == 0 then
						print("Player '" .. name .. "': " .. PlayMate:getColoredRanking( playerObj["Rank"] ))
					else
						print("Player '" .. name .. "-" .. realm .. "': " .. PlayMate:getColoredRanking( playerObj["Rank"] ))
					end
					
					-- Toggle the flag
					printed = true
				end
				
				-- Add this member to the ignore/party list
				PlayMate.Party[ name ] = true
			end
		end
	else
		-- Both are zero, we left the party
		PlayMate.Party = {}
	end
end

-- ================================================================
-- ==                          Network                           ==
-- ================================================================

-- This function handles incoming addon messages. We might support other addons like Heroes or DoIKnowU by the time the GUI
-- is done. Meanwhile, the message should hold the command and the data as encoded as follows.
function PlayMate:handleCommunication(prefix,message,type,sender)
	-- Lowercase index value
	sender_l = sender:lower()

	-- Filter out messages we don't understand 
	if prefix == "PlayMate" and sender_l ~= PlayMate.PlayerName:lower() then
		--if PlayMate.debug then print("PlayMate Raw Data: ", message) end
		
		-- Decode the message into a table, cmd holds the command label
		data = PlayMate:decodeTable( message )
		assert( data["cmd"] ~= nil, "Message decode failed: no 'cmd' found in data")
		
		-- Default: set the remote version to an invalid version number
		local remoteVersion = 0
		local protocolVersion = 0
		if data.Protocol ~= nil then
			-- A protocol version was sent along, use it
			protocolVersion = tonumber(data.Protocol)
		end
			
		if data["Version"] ~= nil then
			-- Extract the remote version number (if set)
			remoteVersion = tonumber(data.Version)
			
			-- Use the version number as protocol version if it was not sent by an older version
			if protocolVersion == 0 then
				protocolVersion = remoteVersion
			end
		end
		
		-- Make sure the protocol versions match and ignore the request otherwise since
		-- different versions could send different data.
		if PlayMate.Protocol ~= protocolVersion then
			if PlayMate.debug then 
				print("PlayMate message from " .. sender .. " ignored because of protocol version mismatch: ours is " .. PlayMate.Protocol .. ", theirs is " .. protocolVersion) 
			end
			
			-- Bail out to prevent communication with incompatible versions
			return
		end
		
		-- If the protocol version is the same, check if the addon version might be newer on the remote end
		if remoteVersion > PlayMate.Version then
			print("A newer version of PlayMate detected! You have " .. PlayMate.Version .. " and should upgrade to version " .. data.Version .. ".")
		end
		
		-- Figure out which command is send and if needed, send a response
		if data["cmd"] == "update" then
			-- Safety checks
			assert(data["UUID"] ~= nil, "PlayMate error: a player ranking was received from " .. sender .. " without a user ID")

			-- Perhaps the version warning should be printed again?
			-- print("|cFFFF0000ERROR: |r" .. sender .. " has sent a ranking from an old version of PlayMate; please ask them to upgrade")

			-- Make sure the ranking we received is not for us, if so, ignore it. Otherwise you will show up in
			-- party reports or people will get angry if known people have given them a bad rank.
			-- Check if the ranking is about a character owned by this account
			if PlayMate.Settings.MyCharacters[  data["PlayerName"]:lower() ] ~= nil then
				if PlayMate.debug then print("Received personal ranking from " .. sender .. " - discarded") end
				return
			end
		
			-- This is an update broadcast, just add it to the shared storage after removing the command...
			data["cmd"] = nil
			-- and the UUID
			local UUID = data["UUID"]
			data["UUID"] = nil
			
			-- Test if this sender is known or create the table in the shared storage otherwise
			if PlayMate.Settings.Shared[ UUID ] == nil then
				-- Save the current player name as the name for the account (each account has one UUID)
				PlayMate.Settings.Shared[ UUID ] = {
					["PlayerName"] = sender,	
					["LastSync"] = 0,
					["Data"] = {}
				}
			end

			-- Copy all relevant fields by hand: this fixes potential abuse from remote senders which transmit garbage (or obsolete field etc)
			local record = {}
			
			-- Check for existence of all fields
			assert(data["Created"]~=nil and data["Note"]~=nil and data["PlayerName"]~=nil and data["Realm"]~=nil and data["Modified"]~=nil and data["Rank"]~=nil,
					"PlayMate error: ranking update from " .. sender .. " did not contain all required field - this is a bug!")
			
			-- Copy all fields
			record["Created"]    = tonumber(data["Created"], 10)
			record["Note"]       = data["Note"]:trim()
			record["PlayerName"] = data["PlayerName"]:trim()
			record["Realm"]      = data["Realm"]:trim()
			record["Modified"]   = tonumber(data["Modified"], 10)
			record["Rank"]       = tonumber(data["Rank"], 10)
			
			-- Now do some safety checks on the data
			assert(record["Note"]:len() < 128, "PlayMate error: ranking update from " .. sender .. " contained an invalid note lenght")
			assert(record["PlayerName"]:len() < 64 and record["PlayerName"]:len() > 3, "PlayMate error: ranking update from " .. sender .. " contained an invalid character name")
			assert(record["Realm"]:len() < 32 and record["Realm"]:len() > 5, "PlayMate error: ranking update from " .. sender .. " contained an invalid realm name")
			assert(record["Rank"] >= -5 and record["Rank"] <= 5, "PlayMate error: ranking update from " .. sender .. " contained invalid rank")
			
			-- Test if the realm exists
			if PlayMate.Settings.Shared[ UUID ]["Data"][ record["Realm"]:lower() ] == nil then
				PlayMate.Settings.Shared[ UUID ]["Data"][ record["Realm"]:lower() ] = {}
			end
			
			-- Add to the shared storage for the sender
			PlayMate.Settings.Shared[ UUID ]["Data"][ record["Realm"]:lower() ][ record["PlayerName"]:lower() ] = record
			
			-- Mark the metadata as dirty so it will rebuild asap
			PlayMate.Settings.Dirty = true
			
			-- Inform user of update
			if record["Realm"]:lower() == PlayMate.Realm:lower() then
				print("PlayMate: received ranking from " .. sender .. " for player |c" .. PlayMate.ColorMap["Text1"] .. record["PlayerName"] .. "|r: " .. PlayMate:getColoredRanking( record["Rank"] ) .. " - Note: " .. record["Note"])
			else
				print("PlayMate: received ranking from " .. sender .. " for player |c" .. PlayMate.ColorMap["Text1"] .. record["PlayerName"] .. "-" .. record["Realm"]  .. "|r: " .. PlayMate:getColoredRanking( record["Rank"] ) .. " - Note: " .. record["Note"])
			end

			if PlayMate.debug then print("Player " .. sender .. " has UUID " .. UUID) end
		elseif data["cmd"] == "ping" then
			-- Request from another player to let him know we are running PlayMate as well
			if PlayMate.debug then
				print("Received ping from " .. sender .. " running PlayMate " .. data["Version"])
			end

			tbl = {
				["cmd"] = "pong",
				["Version"] = PlayMate.Version,
				["Protocol"] = PlayMate.Protocol,
				["RankCount"] = PlayMate:countLocalRankings(),
				["UUID"] = PlayMate.Settings["UUID"]
			}
			
			-- Encode the response
			str = PlayMate:encodeTable( tbl )
			
			-- Whisper the response back
			ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", str, "WHISPER", sender)
		elseif data["cmd"] == "pong" then
			-- When a ping is received, see if we actually want to share the rankings with this channel - otherwise just keep quiet
			if PlayMate.Settings.ShareGuild == false and type == "GUILD" then
				if PlayMate.debug then print("Sync request ignored from " .. sender .. " (guild sharing disabled)") end
				return
			end
			if PlayMate.Settings.ShareParty == false and type == "PARTY" then
				if PlayMate.debug then print("Sync request ignored from " .. sender .. " (party sharing disabled)") end
				return
			end

			-- Response to a ping broadcast in guild or party telling us what version they are running and how many
			-- rankings they have in their local database.
			-- We will respond with a 'sync' command requesting any modified ratings since the last sync.
			if PlayMate.debug then
				print("Received pong from " .. sender .. " running PlayMate " .. data["Version"] .. " and holding " .. data["RankCount"] .. " rankings")
			end
			
			-- Make sure a UUID was send along
			assert(data["UUID"] ~= nil, "PlayMate error: pong command does not contain required user ID")
			
			-- Request the sending of all rankings modified after the last sync date. If this sender has not yet sent us any data, this
			-- means a full sync of their local data.
			tbl = {
				["cmd"] = "sync",
				["From"] = 0,
				["Version"] = PlayMate.Version,
				["Protocol"] = PlayMate.Protocol
			}
			
			-- See if we have synced before
			if PlayMate.Settings.Shared[ data["UUID"] ] ~= nil then
				tbl["From"] = PlayMate.Settings.Shared[ data["UUID"] ]["LastSync"]
			end
			
			-- Encode the command
			str = PlayMate:encodeTable( tbl )
			
			if PlayMate.debug then
				print("Requesting sync with " .. sender .. " from timestamp " .. tbl["From"])
			end
			
			-- Whisper the response back - this tells the remote host to send all database updates since the last sync
			ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", str, "WHISPER", sender)
		elseif data["cmd"] == "sync" then
			-- Sync request, if a date is given we only return the updated records since that date
			local date = tonumber(data["From"], 10)
			
			if PlayMate.debug then
				print("Sending all rankings starting from timestamp " .. date .. " to player " .. sender)
			end
			
			for realm, playerList in pairs(PlayMate.Settings.Mine) do
				for playerName, playerObj in pairs(playerList) do
					if tonumber(playerObj["Modified"], 10) > date then
						if PlayMate.debug then
							print("Sending ranking for " .. playerName .. " to " .. sender)
						end
					
						-- Ranking is newer than given date so send it
						PlayMate:broadcastPlayerRanking( playerObj, sender )
					end
				end
			end
			
			-- Signal that we sent all rankings
			tbl = {
				["cmd"] = "sync_complete",
				["Version"] = PlayMate.Version,
				["Protocol"] = PlayMate.Protocol,
				["UUID"] = PlayMate.Settings["UUID"]
			}
			-- Encode it
			str = PlayMate:encodeTable(tbl)
			-- Send it to complete the sync operation
			ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", str, "WHISPER", sender)
		elseif data["cmd"] == "sync_complete" then
			-- Sent after a player has sent all their modified rankings in response to a sync request
			-- We update the lastSync timestamp to now
			if PlayMate.debug then
				print("Player " .. sender .. " has sent all rankings - sync complete")
			end
			
			-- Make sure a UUID was send along
			assert(data["UUID"] ~= nil, "PlayMate error: sync_complete command does not contain required user ID")
				
			if PlayMate.Settings.Shared[ data["UUID"] ] ~= nil then
				PlayMate.Settings.Shared[ data["UUID"] ]["LastSync"] = time()
			end
		else
			-- Unknown command
			print("PlayMate error: received unknown command '" .. cmd .. "' from " .. sender)
		end
	end
end

-- This function takes a player table (for example from Settings.Mine) and broadcasts it to everyone in guild.
-- Used when updating or adding a player to your database to notify others.
-- Type is optional and defaults to GUILD, if a target is given it is forced to WHISPER
function PlayMate:broadcastPlayerRanking( playerObj, target, type )
	assert( playerObj ~= nil, "playerObj is nil in broadcastPlayerRanking" )

	-- Check if a type is given, if not, revert to guild
	if type == nil then
		type = "GUILD"
	end
	
	-- Duplicate table so we can add and modify values
	playerObj = tcopy( playerObj )
	
	-- Start by inserting the version number (since we only sync up with versions running the same protocol)
	playerObj["Version"] = PlayMate.Version
	playerObj["Protocol"] = PlayMate.Protocol
	
	-- Add the UUID from this player account
	playerObj["UUID"] = PlayMate.Settings["UUID"]

	-- Add the command to the table object
	playerObj["cmd"] = "update"
	
	-- If ShareNote is enabled, remove the note
	if playerObj["ShareNote"] == false then
		playerObj["Note"] = ""
	end
	-- Remove the flag whether the note is shared or not
	playerObj["ShareNote"] = nil

	-- Encode the complete table
	data = PlayMate:encodeTable( playerObj )
	
	-- Make sure the data does not exceed the 254 char boundary
	if data:len() > 246 then
		print("PlayMate error: the message length would exceed 255 bytes - ranking NOT shared!")
		return
	end

	-- Send the data to other players
	if target == nil then
		ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", data, type, nil)
	else
		ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", data, "WHISPER", target)
	end
end

-- Ping the guild to request each player to let us know they are online and using PlayMate
-- The response from the ping is used to trigger a sync between each online player and us
function PlayMate:broadcastPing()
	-- Safety: only ping when we are in guild and it is wanted
	if IsInGuild() ~= 1 then
		print("PlayMate: you are not in a guild, skipping broadcast")
		return
	end

	print("PlayMate: broadcasting ping to sync")

	tbl = {
		["cmd"] = "ping",
		["Version"] = PlayMate.Version,
		["Protocol"] = PlayMate.Protocol
	}
	
	-- Encode the complete table
	data = PlayMate:encodeTable( tbl )
	
	-- Send it
	ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", data, "GUILD", nil)
end

-- Broadcast to the guild to let other players know we got online and are ready to be polled for a sync.
-- This means that this command will send a pong to everyone in guild triggering a mass sync to this player
-- (normally the pong is a response to a ping broadcast from a player sent over a whisper instead of a broadcast)
-- This means it is a smart thing to do when logging in and not during play because it might hang us for a second.
function PlayMate:broadcastPong()
	tbl = {
		["cmd"] = "pong",
		["UUID"] = PlayMate.Settings["UUID"],
		["Version"] = PlayMate.Version,
		["Protocol"] = PlayMate.Protocol,
		["RankCount"] = PlayMate:countLocalRankings()
	}
	
	-- Encode the complete table
	data = PlayMate:encodeTable( tbl )
	
	-- Send the pong to everyone
	ChatThrottleLib:SendAddonMessage("NORMAL", "PlayMate", data, "GUILD", nil)
end

-- ================================================================
-- ==                           Hooks                            ==
-- ================================================================

function PlayMate:updateToolTip()
	-- Check if we want a tooltip modification
	if PlayMate.Settings.ShowInTooltip == false then return end

	-- Whenever this fires, a unit tool tip has popped up. Check if the given unit is a player.
	unitLabel, unitId = GameTooltip:GetUnit()
	
	-- Safety check to make sure only on the right events we try to keep going
	--assert(unitLabel ~= nil and unitId ~= nil, "Popup is not a unit type, something went wrong")
	-- Quick switching might trigger it wrongly, simply abort instead of throwing an error
	if unitLabel == nil or unitId == nil then return end

	-- Test if the unit is a player
	if UnitIsPlayer( unitId ) == 1 then
		-- Fetch the name and realm name for the target
		tgtName, tgtRealm = UnitName( unitId )

		-- If the realm name is empty, its a local player
		if tgtRealm == nil or tgtRealm:trim():len() == 0 then tgtRealm = PlayMate.Realm end
	
		-- This is in fact a player, try to look up the name
		local playerObj = PlayMate:searchPlayer( tgtName, tgtRealm )
		
		if playerObj ~= nil then
			-- Player ranking found, add to tooltip
			local voteStr = playerObj["Votes"] .. " vote"
			-- Rewrite as plural if not one vote
			if( tonumber(playerObj["Votes"]) ~= 1) then voteStr = voteStr .. "s" end
			
			-- Add to the tooltip
			GameTooltip:AddLine("Rank: " .. PlayMate:getColoredRanking( playerObj["Rank"] ) .. " - " .. voteStr)
			
			-- Check if we want to display a note (if set) for this player in the tooltip as well
			if PlayMate.Settings.ShowNoteInTooltip == true then
				-- Add a note as well if we have one
				if playerObj["Note"] ~= nil and playerObj["Note"]:len() > 0 then
					GameTooltip:AddLine("Note: |cFFFFFFFF" .. playerObj["Note"] .. "|r")
				end
			end
		end
	end
end
GameTooltip:HookScript("OnTooltipSetUnit", PlayMate.updateToolTip);

function PlayMate:handleEvent(event, arg1, arg2, arg3, arg4)
	if event == "ADDON_LOADED" and arg1:lower() == "playmate" then
		PlayMate:loadSettings()
	elseif event == "PLAYER_LEAVING_WORLD" then
		PlayMate:saveSettings()
	elseif event == "PARTY_MEMBERS_CHANGED" then
		PlayMate:rankParty()
	elseif event == "CHAT_MSG_ADDON" then
		PlayMate:handleCommunication(arg1, arg2, arg3, arg4)
	elseif event == "PLAYER_ENTERING_WORLD" then
		-- Rebuild the metadata when we enter the world if it has changed
		if PlayMate.Settings ~= nil and PlayMate.Settings.Dirty ~= nil and PlayMate.Settings.Dirty then
			PlayMate:rebuildMetadata()
		end
	--else
	--	-- Unknown event, print debug info
	--	print("Event fired: ", event, " arg1: ", arg1, " arg2: ", arg2, " arg3: ", arg3)
	end
end

-- The whisper filter used when a whisper is received
function msgWhisperFilter(self,event,msg,srcPlayerName, ...)
	-- Test if we want this
	if PlayMate.Settings.ShowInWhispers == false then return end
	
	-- Explode the channel name into a name and realm part (if a player is from another realm)
	name, realm = PlayMate:decodeCharacterName(srcPlayerName)

	-- Try to find a player object for this player by name
	playerObj = PlayMate:searchPlayer(name, realm)
  
	-- Test if we found something
	if playerObj == nil then 
		return false
	end

	-- We have something, grab the color for the given rank
	color = PlayMate:getRankColor( playerObj["Rank"] )
	rank = PlayMate:getRankName( playerObj["Rank"] )
  
	-- Modify the message
	msg = "|c" .. color .. "(Rank: " .. rank .. ")|r " .. msg
	return false, msg, srcPlayerName, ...
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER",msgWhisperFilter)

-- The LFG channel filter used when a message is received
function msgChannelFilter(msg, author, language, channel_title, target, flags, zone_id, channel_number, channel_name)
	-- Zone 26 is LFG, abort if this is something else
	if zone_id ~= 26 then return end

	-- Test if we want this
	if PlayMate.Settings.ShowInLFG == false then return end

	-- Try to find a player object for this player by name
	playerObj = PlayMate:searchPlayer(author)
  
	-- Test if we found something
	if playerObj == nil then 
		return false
	end

	-- We have something, grab the color for the given rank
	color = PlayMate:getRankColor( playerObj["Rank"] )
	rank = PlayMate:getRankName( playerObj["Rank"] )
  
	-- Modify the message
	msg = "|c" .. color .. "(Rank: " .. rank .. ")|r " .. msg
	return false, msg, author, language, channel_title, target, flags, zone_id, channel_number, channel_name
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_CHANNEL",msgChannelFilter)

-- ================================================================
-- ==                       Listing Utility                      ==
-- ================================================================

-- Format of entries in PlayMate.Listing.List:
-- Numbered index, from 1 to #PlayMate.Listing.List-1
-- Entry:
--	PlayerName: name of the player in proper formatted case
--  Realm: name of the realm of this player (could be all the same if filtered on realm or multi-realm is disabled)
--  Rank: numeric value of the rank

-- Push all listing functions in a new nested object
PlayMate.Listing = {}
PlayMate.Listing.List = {}
PlayMate.Listing.RealmList = {}

-- Set the default sorting mode
PlayMate.Listing.SortMode = "name"
-- Set the filters
PlayMate.Listing.NameFilter = nil		-- Name filter, works on partial names
PlayMate.Listing.RealmFilter = nil		-- Realm filter, works only on full realm names

-- This function erases the listing for the GUI. It should be called when the listing frame is hidden 
-- to preserve memory for large listings.
function PlayMate.Listing:clear()
	-- Wipe the current internal listing
	PlayMate.Listing.List = {}
	
	-- Wipe the list of realms we have rankings for
	PlayMate.Listing.RealmList = {}
end

-- Compare 2 entries from the Listing list and return if a > b based on the sorting order
function PlayMate.Listing.compare(a, b)
	if a == nil or b == nil or a["PlayerName"] == nil or b["PlayerName"] == nil or 
		a["Realm"] == nil or b["Realm"] == nil or 
		a["Rank"] == nil or b["Rank"] == nil then
		print("PlayMate.Listing.compare: Error while sorting nil values")
		return false
	end
	
	if PlayMate.Listing.SortMode == "rank" then
		return a.Rank < b.Rank
	end
	
	if PlayMate.Listing.SortMode == "realm" then
		if a.Realm < b.Realm then
			return true
		end
		
		if a.Realm > b.Realm then
			return false
		end
		
		-- If realm names match, compare the player name
		return a.PlayerName < b.PlayerName
	end

	-- If we got here, the sortmode is unknown or its "name"
	return a.PlayerName < b.PlayerName
end

-- Build or rebuild the listing to show in the GUI
function PlayMate.Listing:rebuild()
	-- Wipe the current internal listing
	PlayMate.Listing:clear()
	
	-- Use the meta data cache to construct the current listing and filter out any unwanted entries based
	-- on the filters.
	for realmName, playerList in pairs(PlayMate.Settings.Players) do
		-- If the realm filter is active, test if the names match
		if PlayMate.Listing.RealmFilter == nil or 
			(PlayMate.Listing.RealmFilter ~= nil and PlayMate.Listing.RealmFilter:lower() == realmName) then
	
			-- Save the realm name
			PlayMate.Listing.RealmList[ #PlayMate.Listing.RealmList+1 ] = PlayMate:ReCap(realmName)
			
			-- Iterate over the players in this realm
			for playerName, playerObj in pairs(playerList) do
				-- Check if the name filter is active
				if PlayMate.Listing.NameFilter == nil or
					(PlayMate.Listing.NameFilter ~= nil and playerName:lower():find( PlayMate.Listing.NameFilter:lower()) ) then

					-- Copy a subset of the data
					PlayMate.Listing.List[ #PlayMate.Listing.List+1 ] = {
						["PlayerName"] = PlayMate:ReCap( playerObj["PlayerName"] ),
						["Realm"] = PlayMate:ReCap(realmName),
						["Rank"] = playerObj["Rank"]
					}
				end
			end
		end
	end

	-- Sort the listing
	table.sort(PlayMate.Listing.List, PlayMate.Listing.compare)
	
	if PlayMate.debug == true then
		for id, obj in pairs(PlayMate.Listing.List) do
			print(id .. " " .. obj["PlayerName"])
		end
	end
end


-- ================================================================
-- ==                              UI                            ==
-- ================================================================


function PlayMateOptionsFrame_OnLoad(panel)
	panel.name = "PlayMate"
	InterfaceOptions_AddCategory(panel);
end

function PlayMateOptionsFrame_OnShow(panel)
	-- Copy the active options to the panel widgets
	PlayMateOptionsFrame_chkDebug:SetChecked(PlayMate.debug == true)
	PlayMateOptionsFrame_chkShowRaidRankings:SetChecked(PlayMate.Settings.ShowRaidRankings == true)
	PlayMateOptionsFrame_chkShowPartyRankings:SetChecked(PlayMate.Settings.ShowPartyRankings == true)
	PlayMateOptionsFrame_chkShowInTooltip:SetChecked(PlayMate.Settings.ShowInTooltip == true)
	PlayMateOptionsFrame_chkShowNoteInTooltip:SetChecked(PlayMate.Settings.ShowNoteInTooltip == true)
	PlayMateOptionsFrame_chkShowInLFG:SetChecked(PlayMate.Settings.ShowInLFG == true)
	PlayMateOptionsFrame_chkShowInWhispers:SetChecked(PlayMate.Settings.ShowInWhispers == true)
	PlayMateOptionsFrame_chkShareParty:SetChecked(PlayMate.Settings.ShareParty == true)
	PlayMateOptionsFrame_chkShareGuild:SetChecked(PlayMate.Settings.ShareGuild == true)
	PlayMateOptionsFrame_htmlInfo:SetText("<html><body><h1 align='center'>PlayMate " .. PlayMate.Version .. "</h1><p>By Cyberwizzard</p></body></html>")
end

function PlayMateOptionsFrame_OnHide(panel)
	PlayMate.debug = PlayMateOptionsFrame_chkDebug:GetChecked() == 1
	PlayMate.Settings.ShowRaidRankings = PlayMateOptionsFrame_chkShowRaidRankings:GetChecked() == 1
	PlayMate.Settings.ShowPartyRankings = PlayMateOptionsFrame_chkShowPartyRankings:GetChecked() == 1
	PlayMate.Settings.ShowInTooltip = PlayMateOptionsFrame_chkShowInTooltip:GetChecked() == 1
	PlayMate.Settings.ShowNoteInTooltip = PlayMateOptionsFrame_chkShowNoteInTooltip:GetChecked() == 1
	PlayMate.Settings.ShowInLFG = PlayMateOptionsFrame_chkShowInLFG:GetChecked() == 1
	PlayMate.Settings.ShowInWhispers = PlayMateOptionsFrame_chkShowInWhispers:GetChecked() == 1
	PlayMate.Settings.ShareParty = PlayMateOptionsFrame_chkShareParty:GetChecked() == 1
	PlayMate.Settings.ShareGuild = PlayMateOptionsFrame_chkShareGuild:GetChecked() == 1
	PlayMate:saveSettings()
end

-- GUI stuff below
function PlayMate:showAddForm(playerName, ranking, note, share)
	local frame = getglobal("PlayMateAddForm")
	
	-- If we found the frame object then continue
	if frame then
		if frame:IsVisible() then
			-- Already editing something, hide the frame
			frame:Hide()
		end
		
		-- Load the values if given
		if playerName ~= nil then
			getglobal("PlayMateAddForm_txtName"):SetText( playerName )
		else
			getglobal("PlayMateAddForm_txtName"):SetText( "" )
		end
		if ranking ~= nil then 
			-- Convert to number and round it
			ranking = floor(tonumber(ranking))
			-- Set the slider value
			getglobal("PlayMateAddForm_sldRanking"):SetValue( ranking ) 
			getglobal("PlayMateAddForm_lblRankingLabel"):SetText( PlayMate:getColoredRanking( ranking ) )
		else
			-- Set the neutral value
			getglobal("PlayMateAddForm_sldRanking"):SetValue(0)
			getglobal("PlayMateAddForm_lblRankingLabel"):SetText( PlayMate:getColoredRanking( 0 ) )
		end
		
		if note ~= nil then
			getglobal("PlayMateAddForm_txtNote"):SetText( note )
		else
			getglobal("PlayMateAddForm_txtNote"):SetText( "" )
		end
		
		-- Set the flag indicating if the note should be shared
		-- Note that a nil value will be defaulted to checked
		if share == nil then share = true end
		getglobal("PlayMateAddForm_chkShareNote"):SetChecked( share == true )
		
		-- Show the frame
		frame:Show()
	else
		print("Form for adding not found - this might be a bug!")
	end
end

function PlayMate:doFormSave(btn)
	if btn == nil then
		print("No frame passed")
		return
	end
	
	-- Generate the lowercased index value
	local name = getglobal("PlayMateAddForm_txtName"):GetText():lower()
	
	-- Check if the ranking is about a character owned by this account
	if PlayMate.Settings.MyCharacters[ name ] ~= nil then
		print("PlayMate: you are not allowed to rank your own characters (for obvious reasons) - ranking denied")
		return
	end
	
	-- Add player to the local storage
	PlayMate:addPlayer(
		getglobal("PlayMateAddForm_txtName"):GetText(),
		getglobal("PlayMateAddForm_sldRanking"):GetValue(),
		getglobal("PlayMateAddForm_txtNote"):GetText(),
		getglobal("PlayMateAddForm_chkShareNote"):GetChecked() == 1,
		true
	)
	
	-- Hide the frame
	btn:GetParent():Hide() 
end

function PlayMate_UnitPopupMenu_Click(menu)
	-- Test if the ADDRANK menu item we added was activated
	if menu.value == "ADDRANK" then
		local dropdownFrame = getglobal("UIDROPDOWNMENU_INIT_MENU");
		
		-- Make sure we have something
		assert(dropdownFrame ~= nil, "PlayMate error: dropdown context menu not found!")
		
		-- Fetch the unit from which the menu originated
		local unit = dropdownFrame.unit
		local name = dropdownFrame.name
		local realm = ""
		
		-- If the unit name is nil, the call originated from a chat frame
		if unit == nil then
			name, realm = PlayMate:decodeCharacterName(name)
		else
			-- Fetch all info from this unit
			name, realm = UnitName(unit)
		end
		
		local has_realm = false
		
		if realm == nil or realm:trim():len() == 0 then
			print("PlayMate: ranking player '" .. name .. "'")
			-- Make sure a realm name is set
			realm = PlayMate.Realm
		else
			print("PlayMate: ranking player '" .. name .. "' from '" .. realm .. "'")
			has_realm = true
		end

		-- Safety check
		assert(name ~= nil, "PlayMate error: Popup menu triggered on something else than a player")
		
		name_l  = name:lower()
		realm_l = realm:lower()

		-- This is a player, start the ranking GUI with the name
		-- First look em up
		if PlayMate.Settings.Mine[realm_l] ~= nil and PlayMate.Settings.Mine[realm_l][name_l] ~= nil then
			if PlayMate.debug then print("Updating existing rank for player") end
		
			-- Fetch the object for this player
			local playerObj = PlayMate.Settings.Mine[realm_l][name_l]
			-- If its a cross realm target, use the proper naming
			if has_realm then
				PlayMate:showAddForm(name .. "-" .. realm, playerObj["Rank"], playerObj["Note"], playerObj["ShareNote"] == true )
			else
				PlayMate:showAddForm(name, playerObj["Rank"], playerObj["Note"], playerObj["ShareNote"] == true )
			end
		else
			-- Add as new entry
			if has_realm then
				PlayMate:showAddForm(name .. "-" .. realm)
			else
				PlayMate:showAddForm(name)
			end
		end
	end
end

-- New stuff below

-- Show the main PlayMate UI and select the listings tab as default
function PlayMate:ShowUI()
	-- Generate the master frame holding the PlayMate listing GUI
	local ui = getglobal("PlayMateFrame")
	-- Switch tab button
	PanelTemplates_Tab_OnClick(PlayMateTabHeaderTab1, PlayMateTabHeader);
	-- Update UI to show correct tab
	PlayMateFrame_Update();
	
	ui:Show()
end

-- Show the main GUI and select the usage panel
function PlayMate:ShowUsageUI()
	-- Generate the master frame holding the PlayMate listing GUI
	local ui = getglobal("PlayMateFrame");
	-- Switch tab button
	PanelTemplates_Tab_OnClick(PlayMateTabHeaderTab2, PlayMateTabHeader);
	-- Update UI to show correct tab
	PlayMateFrame_Update();
	
	-- Show main frame if its not yet shown
	if ui then
		--if ui:IsVisible() then
		--	-- Hide it so we can show it again
		--	ui:Hide()
		--end
		
		-- Show the frame
		ui:Show();
	else
		print("PlayMate UI not found - this might be a bug!")
	end
end

PLAYMATEFRAME_SUBFRAMES = { "PlayMateListFrame", "PlayMateUsageFrame", "PlayMateAboutFrame" };
function PlayMateFrame_ShowSubFrame(frameName)
	for index, value in pairs(PLAYMATEFRAME_SUBFRAMES) do
		if ( value == frameName ) then
			_G[value]:Show()
		else
			_G[value]:Hide();
		end	
	end 
end

-- This function updates the main PlayMate UI and shows the currently active tab
function PlayMateFrame_Update()
	if ( PlayMateTabHeader.selectedTab == 1 ) then
		--ShowFriends();
		PlayMateFrameTitleText:SetText("PlayMate - Rankings");
		PlayMateFrame_ShowSubFrame("PlayMateListFrame");
	elseif ( PlayMateTabHeader.selectedTab == 2 ) then
		PlayMateFrameTitleText:SetText("PlayMate - Usage");
		PlayMateFrame_ShowSubFrame("PlayMateUsageFrame");
	elseif ( PlayMateTabHeader.selectedTab == 3 ) then
		PlayMateFrameTitleText:SetText("PlayMate - About");
		PlayMateFrame_ShowSubFrame("PlayMateAboutFrame");
	end
end

-- This function is called when the main PlayMate UI frame is shown - it prepares the ranking listing
function PlayMateFrame_OnShow()
	-- Show the scroll pane
	--PlayMateScrollFrame:Show()
	
	-- Build the list of player rankings
	PlayMate.Listing:rebuild()
	
	-- Update the viewport
	if PlayMate.debug then
		print("Player count " .. #PlayMate.Listing.List)
	end
	-- Update the faux scroll listing, whether its visible or not is irrelevant
	FauxScrollFrame_Update(PlayMateScrollFrame,#PlayMate.Listing.List,7,44);
	
	-- Trigger the rendering of the listing
	PlayMateScrollFrame_Update()
end

-- This function is called when the main PlayMate UI frame is hidden to clean up
function PlayMateFrame_OnHide()
	-- Wipe the internal ranking list for the gui to save memory
	PlayMate.Listing:clear()
end

-- Called each time the scroll frame changed its viewport (scolled up- or down), can also be called when the list is shown
function PlayMateScrollFrame_Update()
	-- PlayMate.PlayerCount is max entries, 7 is number of lines, 44 is pixel height of each line
	FauxScrollFrame_Update(PlayMateScrollFrame,#PlayMate.Listing.List,7,44);
	
	local offset = FauxScrollFrame_GetOffset(PlayMateScrollFrame);
	--print("We're at "..offset);
	
	for i=1,7 do
		local obj = PlayMate.Listing.List[i+offset];
		local f = getglobal("PlayMateEntry"..i);
		--f.BtnName:SetText("View")
		
		if obj ~= nil then
			f.TxtName:SetText(obj.PlayerName);
			f.TxtRealm:SetText(obj.Realm);
			f.TxtScore:SetText(obj.Rank);
			
			local col = PlayMate.HexMap[tostring(obj.Rank)];
			if col == nil then
				print("PlayMate error: can not get color for rank " .. obj.Rank);
			else
				f.TxtScore:SetTextColor(col[1], col[2], col[3], 1.0);
			end
			
		else
			f.TxtName:SetText("");
			f.TxtRealm:SetText("");
			f.TxtScore:SetText("");
		end
	end
end

function PlayMateUsageFrameScrollFrame_AdjustScroll(offset)

end

-- Utility function: capitalize the first letter of each word
function PlayMate:ReCap(str)
	local newstr = ""
	for word in str:gmatch("%w+") do
		newstr = newstr .. word:gsub("^%l", string.upper) .. " "
	end
	return newstr:trim()
end

-- Create the context menu item
UnitPopupButtons["ADDRANK"] = { text = "Rank Player", dist = 0, space = 1 }
-- Party portrait
UnitPopupMenus["PARTY"][ getn(UnitPopupMenus["PARTY"]) ] = "ADDRANK"
-- Chat channel player names (including trade, raid, party etc)
UnitPopupMenus["FRIEND"][ getn(UnitPopupMenus["FRIEND"]) ] = "ADDRANK"
-- Random players
UnitPopupMenus["PLAYER"][ getn(UnitPopupMenus["PLAYER"]) ] = "ADDRANK"

hooksecurefunc("UnitPopup_OnClick", PlayMate_UnitPopupMenu_Click);

-- Start the addon after loading all code
PlayMate:Init()